# Generated by Django 3.0.4 on 2020-03-04 11:11
from __future__ import annotations

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0004_add_authorization_fields"),
    ]

    operations = [
        migrations.CreateModel(
            name="DatasetField",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                ("auth", models.CharField(blank=True, max_length=250, null=True)),
                (
                    "table",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="fields",
                        to="datasets.DatasetTable",
                    ),
                ),
            ],
            options={
                "verbose_name": "Dataset Field",
                "verbose_name_plural": "Dataset Fields",
                "ordering": ("name",),
                "unique_together": {("table", "name")},
            },
        ),
    ]
